function [ILas, IDas, xj, Las, errest] = compute_integral_factors_LDLT(LQ, DQ, h, as, args, ts)
% Compute an approximation to the low-rank factorizations LDL^T of the 
% integrals 
%
% \int_{0}^{a*h}{ exp(s(A M^{-1})^T)M^{-T} LQ DQ LQ^T M^{-1} exp(s A M^{-1}) ds}.
%
% for all the values of a in the vector as by quadrature. Also handles the 
% time-varying case, where the arguments of the exponentials are integrals,
% and the outer integration interval is [t, t+h]. In this case, specify the
% desired values of t in the final parameter ts. The integral is then 
% approximated over [ts(1), ts(1)+h], ..., [ts(end), ts(end)+h], where h is
% scaled by as(1), ..., as(end). The resulting output is a two-level 
% structure with ILas{i}{k} and IDas{i}{k} corresponding to ts(i) and 
% as(k). This is the case even if ts is not specified, in this case use
% ILas{1} to get the desired data.
% 
% Parameters for the approximation are supplied in the structure 
% args.quadrature
% 
% quadrature.type can be either 'gauss', 'clenshawcurtis' or 'equidistant'.
% 'gauss' means Gauss quadrature, which has optimal order for its number of
% points, but with no error estimate available.
% 'clenshawcurtis' uses a Clenshaw-Curtis formula
% 'adaptive' also uses a Clenshaw-Curtis formula, where the nodes may be 
% reused for error estimation. This is then used for increasing the 
% quadrature order until the requested tolerance quadrature.tol is met.
% The starting order is fixed to 4 and ignores quadrature.order.
% 'equidistant' uses equidistantly placed quadrature nodes and computes the
% corresponding weights by solving a Vandermonde system.
% If adaptive time-stepping is used with the option to recompute the
% integral factorization, equidistant nodes are always used and this option
% is ignored.
%
% quadrature.order specifies the quadrature order (default 29, which might 
% be slightly overkill).
% 
% 
% For the time-varying case, quadrature.blocking may be set to true in
% order to compute all the factors at the quadrature nodes simultaneously.
% This will typically be faster due to BLAS-3 blocking effects, but will 
% increase the memory requirements significantly.
% If this option is set, also quadrature.LQ_size needs to be specified. It
% should be the result of size(LQ(t)) for any t.
% 
%
% Output: ILas, IDas such that ILas{i}{k}IDas{i}{k}ILas{i}{k}' approximates 
%         the integral over [ts(i), ts(i) + as(k)*h]
%         Also returns the quadrature points xj and the list of L-factors 
%         Las{j} that are used to form ILas{j} (only useful (and makes
%         sense) in autonomous case for now.)

%
% Copyright 2018 Tony Stillfjord
%
% This file is part of DREsplit.
% 
% DREsplit is free software: you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation, either version 3 of the License, or
% (at your option) any later version.
% 
% DREsplit is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% 
% You should have received a copy of the GNU General Public License
% along with DREsplit.  If not, see <http://www.gnu.org/licenses/>.
%

    if ~isfield(args, 'quadrature')
        error('Must define structure with parameters for quadrature approximation!')
    end
    
    if ~isfield(args.quadrature, 'order')
        order = 29; % Rather excessive... ?
    else
        order = args.quadrature.order;
    end
    
    adaptive = false;
    if ~isfield(args.quadrature, 'type')
        % default Gauss
        [xj, wj] = gauss_quadrature_parameters(h, order);
        warning('Quadrature type not specified, using Gauss quadrature.')
    else
        if strcmp(args.quadrature.type, 'gauss')
            [xj, wj] = gauss_quadrature_parameters(h, order);
        elseif strcmp(args.quadrature.type, 'clenshawcurtis') 
            if rem(order,2) == 1
                error('Quadrature order must be even for Clenshaw-Curtis.')
            end
            [xj, wj] = clenshawcurtis_parameters(0, h, order);
        elseif strcmp(args.quadrature.type, 'adaptive')
            % nodes/weights computed in loop further down
            adaptive = true;
            order = 4;
        elseif strcmp(args.quadrature.type, 'equidistant')
            % Equidistant nodes - used for the adaptive methods
            xj = linspace(0, h, order+1)';
            wj = compute_quadrature_weights(h, xj);
        else
            warning('Specified quadrature type not implemented, using Gauss quadrature.')
            [xj, wj] = gauss_quadrature_parameters(h, order);
        end
    end
    
    TOL = 1e-4;
    if isfield(args.quadrature, 'tol')
        TOL = args.quadrature.tol;
    else
        if adaptive
            warning('Quadrature tolerance unspecified. Using default value 1e-4.')
        end
    end
    
  
    if strcmp(args.impODE.method, 'timevarying')
        timevarying = true;
        t0 = args.impODEobj.t;
        if nargin < 6
            ts = [t0];
        end
    else
        timevarying = false;
        if nargin < 6
            ts = [0];
        end
    end
    
    blocking = false;
    if isfield(args.quadrature, 'blocking') && args.quadrature.blocking
        blocking = true;
        N = args.quadrature.LQ_size(1);
        Nq = args.quadrature.LQ_size(2);
    end
    
    M = args.M;

    ILas = {};
    Las = {}; % Ls for each a
    IDas = {};

    % If autonomous, we have the same function f throughout
    if ~timevarying
        f = @(s, L) args.impODEobj.eval(0, s, M'\L);
    end

    % If ts was not an input, this will be a single-iteration loop
    for i = 1:length(ts)
        t0 = ts(i);
        errest = Inf;
        while errest > TOL % only for the adaptive version
            if adaptive
                [xj, wj] = clenshawcurtis_parameters(0, h, order);
                [~, wj2] = clenshawcurtis_parameters(0, h, order/2);
                % Sort the nodes from 0 to h instead, for iterative computation
                % (weights are symmetric)
                xj = flipud(xj);
            end
            
            
            % For each scaling factor as(k) compute the values Ls(xj)
            for k = 1:length(as)
                if ~timevarying
                    Ls = {};
                    Ls{1} = f(as(k)*xj(1), LQ);
                    for j = 2:length(xj)
                        dx = as(k)*xj(j) - as(k)*xj(j-1);
                        Ltemp = M'*Ls{j-1};
                        Ls{j} = f(dx, Ltemp);
                    end
                else % Timevarying
                    % See comment at end of file for why this is more
                    % complicated than the autonomous case
                    if ~blocking
                        for j = 1:length(xj)
                            t = t0 + as(k)*xj(j);
                            % Solving for X:
%                             Ls{j} = args.impODEobj.eval(t, as(k)*(h-xj(j)), LQ(t));
                            % Solving for Xhat:
                            Ls{j} = args.impODEobj.eval(t, as(k)*(h-xj(j)), M(t)'\LQ(t));
                        end
                    else % Blocking
                        Nx = length(xj);
                        
                        LQ = zeros(N,  Nx*Nq);
                        % Flip the direction of xj so that h - xj is ascending
                        xj2 = flipud(xj);
                        for j = 1:Nx
                            LQ(:, (j-1)*Nq+1 : j*Nq) = args.LQ(xj(j));
                        end
                        t = xj2(end); % first time point
                        y = args.impODEobj.eval(t, h-xj2, LQ);
                        % count backwards to compensate for flipped xj, then we
                        % don't have to mess around with the weights
                        for j = 1:Nx
                            Ls{Nx-j+1} = y(:, (Nx+1)*(Nx-j)*Nq +1: (Nx+1)*(Nx-j)*Nq + Nq);
                        end
                    end
                end
                Las{i}{k} = Ls;
                ILas{i}{k} = cell2mat(Ls);
                IDas{i}{k} = kron(diag(wj * as(k)), DQ); % Block diagonal with weight(j)*as(k)*DQ as blocks
            end
            
            % Column compress approximations
            for k = 1:length(as)
                [ILas{i}{k}, IDas{i}{k}] = column_compress_LDLT(ILas{i}{k}, IDas{i}{k}, args.CC_tol);
            end
            if adaptive
                % Column compress coarser approximations
                for k = 1:length(as)
                    [ILas2{i}{k}, IDas2{i}{k}] = column_compress_LDLT(cell2mat(Las{i}{k}(1:2:end)), kron(diag(wj2 * as(k)), DQ), args.CC_tol);
                end
                errest = 0;
                for k = 1:length(as)
                    errest = max(errest, outerfrobnormdiff_LDLT(ILas{i}{k}, IDas{i}{k}, ILas2{i}{k}, IDas2{i}{k}));
                end
            else
                break
            end
            order = order * 2;
        end
    end
    
end



% Unlike in the autonomous case, we can no longer reuse already computed 
% exponentials to decrease the computational effort in the time-varying 
% case, since the inner integrals are now applied to different arguments:
% LQ(x1), LQ(x2), ..., LQ(x(N)) instead of just LQ. We could compute 
% exp(int_{t0+x_{N-1}}^{t0+h} f(s) ds LQ(x_{N-1}))
%   = exp(int_{t0+x_{N-1}}^{t0+x_N} f(s) ds LQ(x_{N-1}))
%       + exp(int_{t0+x_{N}}^{t0+h} f(s) ds LQ(x_{N-1})),
% but the last term is not what we compute in the N:th step:
% exp(int_{t0+x_N}^{t0+h} f(s) ds LQ(x_N))
% We would then have to also compute exp(int_{t0+x_N}^{t0+h} f(s) ds * ...
% (LQ(x_{N-1} - LQ(x_N)), i.e. two integrals over a shorter interval 
% instead of one over a longer.
% For the (N-2):th term, we would additionally need the both the Nth and 
% (N-1)th integrals for LQ(x_{N-2}) - LQ(x_{N-1}), as well as the (N-2)th
% integral for the new value LQ(x_{N-2})

% In summary, the procedure for the (N-j)th term would result in j 
% integrals over an interval of roughly size h/N/j (if equidistant nodes)
% rather than one integral over an interval of size h/N. The extra 
% implementation effort for this approach therefore does not seem 
% worthwhile.

% However, we could instead do this: solve for all of the initial 
% conditions LQ(xj) at once and utilize more BLAS-3 level operations. 
% Since the output is needed at different time points h - xj, the amount of
% storage is multiplied by a factor equal to the number of quadrature 
% nodes. Since LQ typically has very few columns, it is likely affordable,
% but if not this memory requirement might be unfeasible. Hence the user 
% has to enable this option explicitly. If so, also the number of columns 
% has to be specified for efficiency reasons. (We can't just check unless 
% we evaluate LQ(t) for some t, and we don't want to do that if this 
% function is called in every time step.)
